/*
 *  @(#)WarpSet.java 1.6 00/12/19 15:43:14
 *
 * Copyright (c) 2000-2001 Sun Microsystems, Inc. All Rights Reserved.
 */
package com.sun.j3d.demos.j3dfly.warp;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Iterator;
import javax.swing.JDialog;
import javax.swing.JOptionPane;
import javax.swing.JComboBox;

import javax.media.j3d.TransformGroup;
import javax.media.j3d.Transform3D;

/**
 * Encapsulates a set of WarpPositions
 *
 * @author Paul Byrne
 * @version	1.6, 12/19/00
 */
public class WarpSet extends Object implements Serializable {

    private static final long serialVersionUID=4278610821202427778L;
    
    private HashMap warps;      // Map of warps indexed by name
    private String defaultWarp;
    private transient JComboBox comboBox;
    
    /** Creates new WarpSet */
    public WarpSet( JComboBox comboBox ) {
        warps = new HashMap();
        this.comboBox = comboBox;
    }
    
    /**
     * Add a new Warp to the set.
     * @param name Name of the warp
     * @param description A description of the Warp (can be null)
     * @param position The position of the warp
     */
    public void addWarpPosition( String name, String description, Transform3D position ) {
        WarpPosition w = new WarpPosition( name, description, position );
        warps.put( name, w );
        comboBox.addItem( name );
    }
    
    /**
     * Remove the named warp
     */
    public void deleteWarpPosition( String name ) {
        warps.remove( name );
        comboBox.removeItem( name );
    }
    
    /**
     * Change the position of the named warp, if the warp does not exist it
     * will be created.
     *
     * @param name Name of the warp
     * @param newPosition The new position for the warp
     */
    public void changeWarpPosition( String name, Transform3D newPosition ) {
        WarpPosition w = (WarpPosition)warps.get( name );
        if (w==null) 
            addWarpPosition( name, null, newPosition );
        else 
            w.setPosition( newPosition );
    }
    
    /**
     * Returns true if name exists in the set
     */
    public boolean nameExists( String name ) {
        return warps.containsKey( name );
    }
    
    /**
     * Clear the warp set and update the combobox
     */
    public void clear() {
        warps.clear();
        comboBox.removeAllItems();
        comboBox.addItem("");
    }
    
    /**
     * Set the viewTG to the position of the named warp.
     *
     * If the named warp does not exist the transformGroup will not
     * be modified.
     *
     * @param name Name of the warp
     * @param viewTG The View TransformGroup
     */
    public void warpTo( String name, TransformGroup viewTG ) {
        WarpPosition w = (WarpPosition)warps.get( name );
        if (w!=null) {
            viewTG.setTransform( w.getPosition() );
        }
    }
    
    /**
     * Create a warp for the supplied position.
     * Present the user with a dialog to obtain the
     * name and description of the warp.
     *
     * @param position The position of the warp
     * @param parent The component which should be the parent of the dialog
     * @return Returns the name of the new warp, or null if user cancelled creation
     */
    public String createWarp( Transform3D position, JDialog parent ) {
        CreateWarpDialog dialog = new CreateWarpDialog( parent, true, this );
        dialog.setVisible( true );
        
        if (dialog.isValid()) {
            addWarpPosition( dialog.getName(), dialog.getDescription(), position );
            return dialog.getName();
        }
        
        return null;
        
    }
    
    public void setDefaultWarp( String name ) {
        defaultWarp = name;
    }
    
    public String getDefaultWarp() {
        return defaultWarp;
    }
    
    /**
     * Returns an array containing all the warp names
     */
    public String[] getNames() {
        return (String[])warps.keySet().toArray(new String[warps.size()]);
    }
    
    java.util.Collection getValues() {
        return warps.values();
    }
    
    /** 
     * Merge set with this warp set
     * If this defaultWarp==null then set it to set.defaultWarp
     *
     * TODO - Handle duplicate names correctly
     */
    public void merge( WarpSet set ) {
       java.util.Collection collection = set.getValues();
       
       Iterator it = collection.iterator();
       while( it.hasNext() ) {
           WarpPosition w = (WarpPosition)it.next();
           addWarpPosition( w.getName(), w.getDescription(), w.getPosition() );
       }
       
       if (defaultWarp==null)
           defaultWarp = set.getDefaultWarp();
    }
    
}
