/*
 *  @(#)PerfBehavior.java 1.6 00/12/19 15:44:54
 *
 * Copyright (c) 2000-2001 Sun Microsystems, Inc. All Rights Reserved.
 */
package com.sun.j3d.demos.utils.behaviors;

import javax.media.j3d.*;
import javax.vecmath.*;

/**
 * A performance measurment behavior. The behavior automatically calibrates
 * itself so that the work in measuring performance does not itself impact 
 * performance. 
 *
 * Performance information will be passed to a registered PerfBehaviorListener,
 * if there is no listener then the frames per second will be printed to 
 * System.out
 *
 * The com.sun.j3d.demos.utils.gui.StatsDialog class  provides a 
 * GUI Dialog which uses this behavior to display performance stats.
 *
 * @author Chien Yang, Paul Byrne
 * @version	1.6, 12/19/00
 */
public class PerfBehavior extends Behavior {
    WakeupOnElapsedFrames FPSwakeup = new WakeupOnElapsedFrames(0);
    private static final int numFramesToDiscard = 8;

    private boolean doCalibration = true;
    private int discardedFrames = 0;
    private int numframes = 0;
    private int maxframes = 1;
    private long currtime = 0;
    private long lasttime = 0;
    private long deltatime;
    private boolean finiteLoop = false;
    private int loop = 0;
    private int loopCount = 6;
    
    private PerfBehaviorListener listener = null;
    
    /**
     * Construct the Behavior
     * The Scheduling bounds are set to inifinity
     */
    public PerfBehavior() {
	setSchedulingBounds(new BoundingSphere(new Point3d(0,0,0),Float.MAX_VALUE));
	setEnable(true);
    }
    
    /**
     * Calibrate the behavior
     */
    public void calibrate() {
        doCalibration = true;
        discardedFrames = 0;
        numframes=0;
        maxframes=1;
        currtime = 0;
        lasttime = 0;
        loop = 0;
        loopCount = 6;
    }
    
    /**
     * Called by Java3D to initialize the behavior
     */
    public void initialize() {
	// Set the trigger for the interpolator
	wakeupOn(FPSwakeup);
    }
    
    /**
     * Add the listener which will be notified whenever the performance
     * stats are updated.
     *
     * Current implementation only supports a single listener
     */
    public void addPerfBehaviorListener( PerfBehaviorListener listener ) {
        if (this.listener!=null) {
            throw new RuntimeException("Implementation only supports a single PerfListener");
        }
        this.listener = listener;
    }
    
    /**
     * Remove the PerfBehaviorListener
     */
    public void removePerfBehaviorListener( PerfBehaviorListener listener ) {
        if (this.listener == listener)
            this.listener = null;
        else
            throw new RuntimeException("Trying to Remove listener which is not current");
    }

    // Called every time the behavior is activated
    public void processStimulus(java.util.Enumeration critera) {
	// Apply Calibration Algorithm :
	// To determine maxframes to run before sampling the time
	// to determine frames per second.
	// testduration = 10000 ,To run test pass for 10 seconds
        
	if (doCalibration) { // do the calibration
	    // Always throw away the next "N" frames *after* the first, dummy
	    // call to postSwap
	    if (discardedFrames <= numFramesToDiscard) {
		if (false) {
		    if (discardedFrames == 0)
			System.out.println("postSwap: initial call discarded");
		    else
			System.out.println("postSwap: frame #" +
					   discardedFrames + " discarded");
		}

		discardedFrames += 1;
		lasttime = System.currentTimeMillis();
	    }
	    else {
		numframes += 1;
		//System.out.println(maxframes+" "+numframes+" "+lasttime+" ");
		if (numframes >= maxframes) {
		    currtime = System.currentTimeMillis();
		    deltatime = currtime - lasttime;
		    //System.out.println("deltatime = " + deltatime +
		    //		          ", numframes = " + numframes);

                    if (deltatime > 1000 && maxframes >= 8 ) {
			maxframes =
			    (int)Math.ceil((double)maxframes * 1000 /
					    (double)deltatime);
			//System.out.println("maxframes = " + maxframes);

			// reset the value for the measurement
			doCalibration = false;
			numframes = 0;
			lasttime = System.currentTimeMillis();
		    }
		    else {
			maxframes *= 2;
		    }
		}
	    }
	}
	else { // do the measurement
	    numframes += 1;
	    if (numframes >= maxframes) {
		currtime = System.currentTimeMillis();
		deltatime = currtime - lasttime;
		double fps = (double)numframes / ((double)deltatime / 1000.0d );
                //System.out.println( numframes+"  "+(deltatime/1000.0)+"  "+fps);
                if (listener!=null) {
                    listener.updatePerformanceFigures( (float)fps );
                } else {
                    System.out.println("fps : " + fps);
                }
		if (finiteLoop) {
		    loop ++;
		    if (loop >= loopCount) System.exit(0);
		}
		lasttime = System.currentTimeMillis();
                numframes = 0;
	    }
	}

	// Set the trigger for the interpolator
	wakeupOn(FPSwakeup);
	
    }
}


