/*
 *  @(#)SceneGraphFileWriter.java 1.5 00/12/19 17:21:59
 *
 * Copyright (c) 2000-2001 Sun Microsystems, Inc. All Rights Reserved.
 */
package com.sun.j3d.demos.utils.scenegraph.io;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;

import javax.media.j3d.VirtualUniverse;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.SceneGraphObject;
import javax.media.j3d.CapabilityNotSetException;
import com.sun.j3d.demos.utils.scenegraph.io.retained.RandomAccessFileControl;

/**
 * Write a (set) of Java3D BranchGraphs and/or Universe to a file. The BranchGraphs 
 * are stored in the order in which they are written, they can be read in any order
 * using SceneGraphFileReader.
 *
 * The API handles Nodes and NodeComponents that are shared between seperate 
 * graphs. It will handle all Java3D 1.2.1 core classes and any user
 * subclass of a Node or NodeComponent that implements the SceneGraphIO
 * interface.
 *
 * @author Paul Byrne
 * @version	1.5, 12/19/00
 */
public class SceneGraphFileWriter extends java.lang.Object {

    private RandomAccessFileControl fileControl;
    private File file;
    
    /** Creates new SceneGraphFileWriter and opens the file for writing
     *
     * <P>Writes the 
     * Java3D Universe structure to the file. This includes the number and position of
     * the Locales, PlatformGeometry, number of MultitransformGroups. However this
     * call does not write the content of the branch graphs unless writeUniverseContent is true
     * This call will overwrite any existing universe, fileDescription and
     * userData in the file.</P>
     * 
     * <P>close() MUST be called when IO is complete. If close() is not called
     * the file contents will be undefined.</P>
     *
     * @param file The file to write the data to
     * @param universe The VirtualUniverse to write
     * @param writeUniverseContent If true the content of the Locales will be written
     * otherwise just the universe configuration data will be written
     * @param fileDescription A description of the files content
     * @param fileUserData User defined object
     *
     * @exception IOException Thrown if there are any IO errors
     * @exception UnsupportedUniverseException Thrown is <code>universe</code> is not
     * a supported universe class. Currently SimpleUniverse is supported.
      */
    public SceneGraphFileWriter( java.io.File file,
                       VirtualUniverse universe,
                       boolean writeUniverseContent,
                       String fileDescription,
                       java.io.Serializable fileUserData) throws IOException, UnsupportedUniverseException {
        fileControl = new RandomAccessFileControl();
        this.file = file;
        file.createNewFile();
        
        if (!file.canWrite())
            throw new IOException( "Can not Write to File" );
        
        fileControl.createFile( file, universe, writeUniverseContent, fileDescription, fileUserData );
    }
    
    /**
     * Write the graph to the end of the file.
     *
     * close() MUST be called when IO is complete. If close() is not called
     * the file contents will be undefined.
     */
    public void writeBranchGraph( BranchGroup graph ) throws IOException {
        writeBranchGraph( graph, null );
    }
    
    /**
     * Write a branch graph and some user associated data to the end of the file
     *
     * close() MUST be called when IO is complete. If close() is not called
     * the file contents will be undefined.
     */
    public void writeBranchGraph( BranchGroup graph,
                                  java.io.Serializable data ) throws IOException {
        fileControl.writeBranchGraph( graph, data );
    }
    
    /**
     * Add a named reference to a SceneGraphObject in the file
     *
     * <code>object</code> must have been written to the file before this method is
     * called. If the object is not in the file a NamedObjectException will be thrown.
     *
     * Adding duplicate names will result in the old name being overwritten.
     * Different names can reference the same object
     */
    public void addObjectName( String name, SceneGraphObject object ) throws NamedObjectException {
        fileControl.addNamedObject( name, object );
    }
    
    /**
     * Close the file and cleanup internal data structures.
     */
    public void close() throws IOException {
        fileControl.close();
    }
    
}
