/*
 *  @(#)RandomAccessFileControl.java 1.9 01/03/02 14:04:25
 *
 * Copyright (c) 2000-2001 Sun Microsystems, Inc. All Rights Reserved.
 */
package com.sun.j3d.demos.utils.scenegraph.io.retained;

import java.io.RandomAccessFile;
import java.io.IOException;
import java.io.DataOutput;
import java.io.DataInput;

import javax.media.j3d.VirtualUniverse;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.CapabilityNotSetException;

import com.sun.j3d.demos.utils.scenegraph.io.state.javax.media.j3d.SceneGraphObjectState;
import com.sun.j3d.demos.utils.scenegraph.io.state.javax.media.j3d.NodeComponentState;
import com.sun.j3d.demos.utils.scenegraph.io.state.javax.media.j3d.BranchGroupState;
import com.sun.j3d.demos.utils.scenegraph.io.UnsupportedUniverseException;

/**
 * @author Paul Byrne
 * @version	1.9, 03/02/01
 */
public class RandomAccessFileControl extends Controller {
        
    private long user_data;
    private long universe_config;
    
    private long symbol_table;
    
    private RandomAccessFile raf;
    
    private int branchGraphCount=0;
    
    private boolean writeMode = false;
    private Object userData;
    private int currentFileVersion;

    /** Creates new RandomAccessFileControl */
    public RandomAccessFileControl() {
        super();
        symbolTable = new SymbolTable(this);
    }
    
    /**
     * Create the file and write the inital header information
     */
    public void createFile( java.io.File file,
                            VirtualUniverse universe,
                            boolean writeUniverseContent,
                            String description,
                            java.io.Serializable userData ) throws IOException,
                                                            UnsupportedUniverseException,
                                                            CapabilityNotSetException {
                       
        raf = new RandomAccessFile( file, "rw" );
        writeMode = true;
        
        raf.seek(0);
        raf.writeUTF( FILE_IDENT );
        
        raf.seek(20);
        raf.writeUTF( FILE_VERSION );
        
        raf.seek( BRANCH_GRAPH_COUNT );
        raf.writeInt( 0 );          // Place holder to branch graph count
        
        raf.seek( FILE_DESCRIPTION );
        
        if (description==null)
            description="";
        raf.writeUTF( description );
        
        writeSerializedData( raf, userData );
        
        universe_config = raf.getFilePointer();
        writeUniverse( raf, universe, writeUniverseContent );
        
    }
    
    /**
     * Open the file for reading
     */
    public void openFile( java.io.File file ) throws IOException {
        raf = new RandomAccessFile( file, "r" );
        writeMode = false;
        
        raf.seek(0);
        String ident = raf.readUTF();
        
        raf.seek(20);
        String version = raf.readUTF();
        
        if (version.equals("0.9"))
            currentFileVersion = 1;
        else if (version.equals("1"))
            currentFileVersion = 2;
        
        // readFileDescription sets user_data
        String description = readFileDescription();
        
        raf.seek( BRANCH_GRAPH_COUNT );
        branchGraphCount = raf.readInt();
        //System.out.println("BranchGraph count : "+branchGraphCount );
        
        raf.seek( UNIVERSE_CONFIG_PTR );
        universe_config = raf.readLong();

        raf.seek( SYMBOL_TABLE_PTR );
        symbol_table = raf.readLong();
        
        VirtualUniverse universe;
        
        raf.seek( symbol_table );
        symbolTable.readTable( raf, false );
        raf.seek(user_data);       

        userData = readSerializedData(raf);
    }
    
    public VirtualUniverse readUniverse( boolean attachBranchGraphs ) throws IOException {
        raf.seek( universe_config );
        return readUniverse( raf, attachBranchGraphs );
    }
    
    public Object getUserData() {
        return userData;
    }
    
    /**
     * Get the file version of the file we are reading
     */
    public int getCurrentFileVersion() {
        return currentFileVersion;
    }
    
    /**
     * Read the set of branchgraps.
     *
     * Used by readUniverse
     *
     * RandomAccessFileControl will read the graphs in the array,
     * StreamControl will read all graphs in the stream
     */
    protected void readBranchGraphs( int[] graphs ) throws IOException {
        for(int i=0; i<graphs.length; i++) {
            readBranchGraph( graphs[i] );
        }
    }
    
    /**
     * Return the number of branchgraphs in the file
     */
    public int getBranchGraphCount() {
        return symbolTable.getBranchGraphCount();
    }
    
    public void writeBranchGraph( BranchGroup bg, java.io.Serializable userData ) throws IOException {
        long filePointer = raf.getFilePointer();
        raf.writeInt( 0 );          // Node count
        writeSerializedData( raf, userData );  // Size and byte[]
        
        //System.out.println("Actual Write at "+raf.getFilePointer() );
        
        SymbolTableData symbol = symbolTable.getSymbol( bg );
        
        if (symbol==null) {
            symbol = symbolTable.createSymbol( bg );
            symbol.branchGraphID = -1;          // This is a new BranchGraph so set the ID to -1
        }                                       // which will cause setBranchGraphRoot to assign a new ID.
        
        symbolTable.setBranchGraphRoot( symbol, filePointer );        
        
        SceneGraphObjectState state = createState( bg, symbol );
        symbolTable.startUnsavedNodeComponentFrame();
        //System.out.println(state);
        try {
            writeObject( raf, state );
            writeNodeComponents( raf );            
        } catch( IOException e ) {
            e.printStackTrace();
        }
        symbolTable.endUnsavedNodeComponentFrame();
        
    }
    
    public BranchGroup[] readBranchGraph( int graphID ) throws IOException {
        //System.out.print("Loading graph "+graphID+" : ");   // TODO - remove
        int[] dependencies = symbolTable.getBranchGraphDependencies(graphID);
        
        //for(int i=0; i<dependencies.length; i++)
        //    System.out.print( dependencies[i]+"  ");        // TODO - remove
        //System.out.println();
        
        BranchGroupState[] states  = new BranchGroupState[ dependencies.length+1];
        BranchGroup[] ret = new BranchGroup[ states.length ];
        states[0] = readSingleBranchGraph( graphID );
                
        for( int i=0; i<dependencies.length; i++) {
            states[i+1] = readSingleBranchGraph( dependencies[i] );
        }
        
        for( int i=0; i<states.length; i++) {
            if (!states[i].getSymbol().graphBuilt) {
                states[i].buildGraph();
                states[i].getSymbol().graphBuilt = true;
            }
            ret[i] = (BranchGroup)states[i].getNode();
        }
        
        symbolTable.clearUnshared();            // Remove all unshared symbols
        
        return ret;
    }
    
    /**
     * Read and return all the graphs in the file
     */
    public BranchGroup[] readAllBranchGraphs() throws IOException {
        int size = getBranchGraphCount();
        BranchGroupState[] states  = new BranchGroupState[ size ];
        BranchGroup[] ret = new BranchGroup[ size ];
                
        for( int i=0; i<size; i++) {
            states[i] = readSingleBranchGraph( i );
        }
        
        for( int i=0; i<states.length; i++) {
            if (!states[i].getSymbol().graphBuilt) {
                states[i].buildGraph();
                states[i].getSymbol().graphBuilt = true;
            }
            ret[i] = (BranchGroup)states[i].getNode();
        }
        
        symbolTable.clearUnshared();            // Remove all unshared symbols
        
        return ret;
    }

    /**
     * Read the specified branchgraph but do NOT call buildGraph
     */
    private BranchGroupState readSingleBranchGraph( int graphID ) throws IOException {
        SymbolTableData symbol = symbolTable.getBranchGraphRoot( graphID );
        
        if (symbol.nodeState!=null) {
            return (BranchGroupState)symbol.nodeState;
        }
        
        raf.seek( symbolTable.getBranchGraphFilePosition( graphID ) );
         
        return readNextBranchGraph();
    }    
    
    /**
     * Read the next userData and BranchGraph structure in the file
     * at the current position
     */
    private BranchGroupState readNextBranchGraph() throws IOException {
        int nodeCount = raf.readInt();
        skipUserData( raf );

        BranchGroupState state=null;
        try {
            state = (BranchGroupState)readObject( raf );
            readNodeComponents( raf );
            
        } catch( IOException e ) {
            e.printStackTrace();
        }
        branchGraphList.add( state );
        
        return state;
    }
        
    public Object readBranchGraphUserData( int graphID ) throws IOException {
        raf.seek( symbolTable.getBranchGraphFilePosition( graphID ) );
        
        int nodeCount = raf.readInt();
        return readSerializedData( raf );        
    }
    
    public void writeObject( DataOutput out, SceneGraphObjectState obj ) throws IOException {
        symbolTable.setFilePosition( raf.getFilePointer(), obj );  
        super.writeObject( out, obj );
    }
        
    public String readFileDescription() throws IOException {
        raf.seek( FILE_DESCRIPTION );
        String ret = raf.readUTF();
        
        user_data = raf.getFilePointer();
        return ret;
    }
    
    /**
     * Used by SymbolTable to load a node component that is not in current
     * graph
     */
    public void loadNodeComponent( SymbolTableData symbol ) throws IOException {
        raf.seek( symbol.filePosition );
        readObject( raf );
    }
    
    /**
     * Loads the specified SharedGroup
     */
    public void loadSharedGroup( SymbolTableData symbol ) throws IOException {
        raf.seek( symbol.filePosition );
        readObject( raf );
    }
            
    public void close() throws IOException {
        if (writeMode)
            writeClose();
        
        raf.close();
        super.reset();
    }
    
    /**
     * Write all the pointers etc
     */
    private void writeClose() throws IOException {       
        symbol_table = raf.getFilePointer();
        super.getSymbolTable().writeTable( raf );
        
        raf.seek( UNIVERSE_CONFIG_PTR );
        raf.writeLong( universe_config );
        raf.seek( SYMBOL_TABLE_PTR );
        raf.writeLong( symbol_table );
        raf.seek( BRANCH_GRAPH_COUNT );
        raf.writeInt( symbolTable.getBranchGraphCount() );
    }
    
    public long getFilePointer() {
        try {
            return raf.getFilePointer();
        }catch(IOException e ) {}
        return 0;
    }

}
