/*
 *  @(#)SceneGraphObjectState.java 1.10 01/03/02 15:34:25
 *
 * Copyright (c) 2000-2001 Sun Microsystems, Inc. All Rights Reserved.
 */
package com.sun.j3d.demos.utils.scenegraph.io.state.javax.media.j3d;

import java.io.*;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Enumeration;
import javax.media.j3d.SceneGraphObject;
import javax.vecmath.Color3f;
import javax.vecmath.Point3d;
import javax.vecmath.Vector4d;
import javax.vecmath.Tuple3d;
import javax.vecmath.Tuple4d;
import com.sun.j3d.demos.utils.scenegraph.io.retained.Controller;
import com.sun.j3d.demos.utils.scenegraph.io.retained.SymbolTableData;

public abstract class SceneGraphObjectState {
    
    protected SceneGraphObject node;
    protected SymbolTableData symbol;
    protected Controller control;
    protected String nodeClassName;
    
    /**
     * Create a new State object
     *
     * During Saveing
     *  SymbolTableData will have the nodeID and j3dNode fields set
     *
     * During loading
     *  SymbolTableData be null, symbol will be created and added to the
     * symbol data during readObject()
     */
    public SceneGraphObjectState( SymbolTableData symbol, Controller control ) {
        this.symbol = symbol;
        this.control = control;
        
        if (symbol!=null) {
            this.node = symbol.j3dNode;
            
            // TODO remove this consistancy check
            if (symbol.j3dNode==null || symbol.nodeID==0)
                throw new RuntimeException( "Bad Symbol in State creation");
        }
        
        
        if (node!=null) {
            nodeClassName = node.getClass().getName();
            
            if (node instanceof com.sun.j3d.demos.utils.scenegraph.io.SceneGraphIO)
                ((com.sun.j3d.demos.utils.scenegraph.io.SceneGraphIO)node).createSceneGraphObjectReferences( control.getSymbolTable() );
        }
        
    }
    
    /**
     * DO NOT call symbolTable.addReference in writeObject as this (may)
     * result in a concurrentModificationException.
     *
     * All references should be created in the constructor
     */
    public void writeObject( DataOutput out ) throws IOException {  
        boolean sgIO = (node instanceof com.sun.j3d.demos.utils.scenegraph.io.SceneGraphIO ) ? true : false;
        out.writeBoolean( sgIO );
        out.writeInt( symbol.nodeID );
        
        
        int nodeClassID = control.getNodeClassID( node );
        
        out.writeShort( nodeClassID );
        
        if (nodeClassID==-1)
            out.writeUTF( nodeClassName );
        
        writeConstructorParams( out );
        
        if (sgIO) {
            ((com.sun.j3d.demos.utils.scenegraph.io.SceneGraphIO)node).writeSceneGraphObject( out );
        }
        
        writeUserData( out );
        
        writeCapabilities( out );
        
    }

    public void readObject( DataInput in ) throws IOException {
       
        boolean sgIO = in.readBoolean();
        int nodeID = in.readInt();
        
        if (control.getCurrentFileVersion()==1) {
            nodeClassName = in.readUTF();
            readConstructorParams( in );
            node = createNode(nodeClassName);
        } else {
            int nodeClassID = in.readShort();

            nodeClassName = null;
            
            if (nodeClassID==-1)
                nodeClassName = in.readUTF();
            
            readConstructorParams( in );
            
            if (nodeClassID!=-1) {
                node = createNode();
                nodeClassName = node.getClass().getName();
            } else 
                node = createNode(nodeClassName);
        }
        

        if (sgIO) {
            ((com.sun.j3d.demos.utils.scenegraph.io.SceneGraphIO)node).readSceneGraphObject( in );
        }
        
           symbol = control.getSymbolTable().createSymbol( this, node, nodeID );
           readUserData( in );

	   readCapabilities( in );
           /*
       if (control.getCurrentFileVersion()>=1) {
       } else
	   throw new RuntimeException( "Unrecognised file record version");
            **/
    }
    
    public SceneGraphObject getNode() {
        return node;
    }
    
    public int getNodeID() {
        return symbol.nodeID;
    }
    
    public SymbolTableData getSymbol() {
        return symbol;
    }
        
    private void readUserData( DataInput in ) throws IOException {
        
        node.setUserData( control.readSerializedData( in ));
    }
    
    private void writeUserData( DataOutput out ) throws IOException {
        Object obj = node.getUserData();
        if (obj != null && !(obj instanceof java.io.Serializable)) {
            System.out.println("UserData is not Serializable and will not be saved");
            obj = null;
        }
        
        control.writeSerializedData( out, (Serializable)obj );
    }
    
    private void writeCapabilities( DataOutput out ) throws IOException {
        int[] capabilities = getCapabilities();
        
        out.writeInt( capabilities.length );
        for( int i=0; i<capabilities.length; i++)
            out.writeInt( capabilities[i] );
    }
    
    private void readCapabilities( DataInput in ) throws IOException {
        int length = in.readInt();
        for( int i=0; i<length; i++)
            node.setCapability( in.readInt() );
    }

    /**
      * Write the parameters required for the constructor of the Java3D object
      */
    protected void writeConstructorParams( DataOutput out ) throws
							IOException {
    }

    /**
      * Read the parameters required for the constructor of the Java3D object
      */
    protected void readConstructorParams( DataInput in ) throws
							IOException {
    }
    
    /**
     * Create a new Java3D node for this object.
     *
     * This method is ONLY used when the Java3D Class type matches the
     * State type, ie this does NOT handle subclasses of Java3D.
     *
     * For Java3D subclasses use createNode( Class state)
     */
    protected abstract SceneGraphObject createNode();
    
    /**
      * Create a new Java3D node from the supplied class using the parameterless constructor
      * 
      * For Java3D nodes which do not have a default constructor you must
      * overload this method and create the object using createNode( className, parameters )
      * This will correctly handle subclasses of Java3D classes
     */
    protected SceneGraphObject createNode( Class state ) {
	SceneGraphObject ret;

	try {
	    ret = (SceneGraphObject)state.newInstance();

	    //System.out.println("Created J3D node for "+className );
	} catch( IllegalAccessException exce ) {
	    throw new RuntimeException( "Broken State class for "+
						state.getClass().getName()+" - IllegalAccess" );
	} catch( InstantiationException excep ) {
	    throw new RuntimeException( "Broken State class for "+
						state.getClass().getName() );
	}

	return ret;
    }    
    
    /**
      * Create a new Java3D node from the supplied class name using the parameterless constructor
      * 
      * For Java3D nodes which do not have a default constructor you must
      * overload this method and create the object using createNode( className, parameters )
      * This will correctly handle subclasses of Java3D classes
     */
    private SceneGraphObject createNode( String className ) {
	SceneGraphObject ret;

	try {
            Class state = Class.forName( className );
            
	    ret = createNode( state );

	    //System.out.println("Created J3D node for "+className );
	} catch(ClassNotFoundException e) {
            if (control.useSuperClassIfNoChildClass())
                ret = createNodeFromSuper( className );
            else
                throw new RuntimeException( "No Such Class "+
						className );
	}

	return ret;
    }
    
    /**
     * If createNode cannot locate the correct class to instantiate
     * the node this method is called and will instantiate the
     * node using it's Java3D Core superclass
     */
    private SceneGraphObject createNodeFromSuper( String className ) {
	SceneGraphObject ret;
        
        String tmp = this.getClass().getName();
        String superClass = tmp.substring( tmp.indexOf("state")+6, tmp.length()-5 );
        
	try {
            Class state = Class.forName( superClass );
            
	    ret = (SceneGraphObject)state.newInstance();

	} catch(ClassNotFoundException e) {
	    throw new RuntimeException( "No Such Class "+
						className );
	} catch( IllegalAccessException exce ) {
	    throw new RuntimeException( "Broken State class for "+
						className+" - IllegalAccess" );
	} catch( InstantiationException excep ) {
	    throw new RuntimeException( "Broken State class for "+
						className );
	}
        
        return ret;
    }
    
    /**
     * Create a Java3D node which does not have a default constructor
     *
     * parameterTypes must contain the classes required by the constructor,
     * use Interger.TYPE, Float.TYPE etc to specifiy primitive types
     *
     * paramters should contain the list of parameters for the constructor,
     * primitive types should be wrapped in the appropriate class (ie Integer, Float )
     */
    private SceneGraphObject createNode( String className, Class[] parameterTypes, Object[] parameters ) {
        SceneGraphObject ret;
        Constructor constructor;

        try {
            Class state = Class.forName( className );
            constructor = state.getConstructor( parameterTypes );
            ret = (SceneGraphObject)constructor.newInstance( parameters );
	} catch(ClassNotFoundException e1) {
            if (control.useSuperClassIfNoChildClass())
                ret = createNodeFromSuper( className, parameterTypes, parameters );
            else
                throw new RuntimeException( "No State class for "+
						className );
	} catch( IllegalAccessException e2 ) {
	    throw new RuntimeException( "Broken State class for "+
						className+" - IllegalAccess" );
	} catch( InstantiationException e3 ) {
	    throw new RuntimeException( "Broken State class for "+
						className );
        } catch( java.lang.reflect.InvocationTargetException e4 ) {
	    throw new RuntimeException( "InvocationTargetException for "+
						className );
        } catch( NoSuchMethodException e5 ) {
            for(int i=0; i<parameterTypes.length; i++)
                System.out.println( parameterTypes[i].getName() );
            System.out.println("------");
	    throw new RuntimeException( "Invalid constructor for "+
						className );
        }
        
        return ret;
    }

    /**
     * Create a Java3D node which does not have a default constructor
     *
     * parameterTypes must contain the classes required by the constructor,
     * use Interger.TYPE, Float.TYPE etc to specifiy primitive types
     *
     * paramters should contain the list of parameters for the constructor,
     * primitive types should be wrapped in the appropriate class (ie Integer, Float )
     */
    protected SceneGraphObject createNode( Class j3dClass, Class[] parameterTypes, Object[] parameters ) {
        SceneGraphObject ret;
        Constructor constructor;

        try {
            constructor = j3dClass.getConstructor( parameterTypes );
            ret = (SceneGraphObject)constructor.newInstance( parameters );
	} catch( IllegalAccessException e2 ) {
	    throw new RuntimeException( "Broken State class for "+
						j3dClass.getClass().getName()+" - IllegalAccess" );
	} catch( InstantiationException e3 ) {
	    throw new RuntimeException( "Broken State class for "+
						j3dClass.getClass().getName() );
        } catch( java.lang.reflect.InvocationTargetException e4 ) {
	    throw new RuntimeException( "InvocationTargetException for "+
						j3dClass.getClass().getName() );
        } catch( NoSuchMethodException e5 ) {
            for(int i=0; i<parameterTypes.length; i++)
                System.out.println( parameterTypes[i].getName() );
            System.out.println("------");
	    throw new RuntimeException( "Invalid constructor for "+
						j3dClass.getClass().getName() );
        }
        
        return ret;
    }
    
    /**
     * If createNode cannot locate the correct class to instantiate
     * the node this method is called and will instantiate the
     * node using it's Java3D Core superclass
     */
    private SceneGraphObject createNodeFromSuper( String className, Class[] parameterTypes, Object[] parameters ) {
	SceneGraphObject ret;
        
        String tmp = this.getClass().getName();
        String superClass = tmp.substring( tmp.indexOf("state")+6, tmp.length()-5 );
        Constructor constructor;

        try {
            Class state = Class.forName( superClass );
            constructor = state.getConstructor( parameterTypes );
            ret = (SceneGraphObject)constructor.newInstance( parameters );
	} catch(ClassNotFoundException e1) {
	    throw new RuntimeException( "No State class for "+
						superClass );
	} catch( IllegalAccessException e2 ) {
	    throw new RuntimeException( "Broken State class for "+
						className+" - IllegalAccess" );
	} catch( InstantiationException e3 ) {
	    throw new RuntimeException( "Broken State class for "+
						className );
        } catch( java.lang.reflect.InvocationTargetException e4 ) {
	    throw new RuntimeException( "InvocationTargetException for "+
						className );
        } catch( NoSuchMethodException e5 ) {
            for(int i=0; i<parameterTypes.length; i++)
                System.out.println( parameterTypes[i].getName() );
            System.out.println("------");
	    throw new RuntimeException( "Invalid constructor for "+
						className );
        }
        
        return ret;
    }

    /**
      * Given a scene graph object instantiate the correct State class
      * for that object
      */
    protected SceneGraphObjectState createState( SceneGraphObject obj, Controller control ) {

        return control.createState( obj );
    }

    /**
      * Return the class name of the Class, the fully qualified classname
      * is stripped of all package information and returned
      */
    private String getClassName( Class c ) {
	return c.getName().substring(c.getName().lastIndexOf('.')+1);
    }

    /**
      * Return a Set of Strings representing all the 
      * capabilities set in this node
      */
    private int[] getCapabilities() {
	ArrayList ret = new ArrayList();

        int value;
        String str;
        Class cl = node.getClass();
	int i=0;

        Field[] fields = cl.getFields();

        try {
            for(i=0; i<fields.length; i++) {
                 str = fields[i].getName();
                 if (str.indexOf("ALLOW") != -1) {
                     value = fields[i].getInt( fields[i] );
                     if ( node.getCapability( value )) {
			 ret.add( new Integer(value) );
			 //System.out.println("CAPABILITY "+str);
		     }
                 }
            }
        } catch(Exception e) {
	    System.out.println( "Field "+fields[i] );
	    System.out.println( "Class "+cl.getName() );
            e.printStackTrace();
            System.exit(1);
        }

        int[] capabilities = new int[ ret.size() ];
        for(int j=0; j<capabilities.length; j++)
            capabilities[j] = ((Integer)ret.get(j)).intValue();

	return capabilities;
    }
    


    public void buildGraph() {
        //System.out.println("Build Graph "+this);
            if (node instanceof com.sun.j3d.demos.utils.scenegraph.io.SceneGraphIO)
                ((com.sun.j3d.demos.utils.scenegraph.io.SceneGraphIO)node).restoreSceneGraphObjectReferences( control.getSymbolTable() );
    }
    
    public void cleanup() {
        control = null;
        node = null;
    }
    
}
