/*
 *  @(#)MouseBehavior.java 1.4 01/02/23 12:44:47
 *
 * Copyright (c) 2000-2001 Sun Microsystems, Inc. All Rights Reserved.
 */
package com.sun.j3d.demos.utils.vpbehaviors;

import java.awt.event.ComponentEvent;
import java.awt.event.MouseEvent;
import java.awt.event.KeyEvent;
import java.awt.AWTEvent;
import java.awt.Component;
import java.awt.Cursor;
import javax.swing.SwingUtilities;

import javax.media.j3d.WakeupOnAWTEvent;
import javax.media.j3d.WakeupOnElapsedFrames;
import javax.media.j3d.WakeupOr;
import javax.media.j3d.WakeupCriterion;
import javax.media.j3d.WakeupCondition;
import javax.media.j3d.TransformGroup;
import javax.media.j3d.Transform3D;
import javax.media.j3d.View;

import javax.vecmath.Vector3f;


/**
 * Abstract class which implements much of the event tracking and
 * state updating shared between Fly and Walk behavior.
 *
 * Canvas3D size changes are tracked
 *
 * @author Paul Byrne
 * @version	1.4, 02/23/01
 */
public abstract class MouseBehavior extends VPBehavior {

    protected WakeupOr awtCondition;

    protected WakeupOr bothCondition;
    protected java.awt.Rectangle canvasBounds;
    protected java.awt.geom.Point2D.Float canvasCenter;
    
    protected Transform3D targetTransform;   
    
    protected float MAX_ANGLE = (float)Math.toRadians( 3 );
    protected float MAX_VELOCITY = 10f;

    protected boolean motion = false;

    protected boolean ignoreMouseMotion = false;
    protected float deadXSize;
    protected float deadYSize;
    protected FlyControlDialog flyControlDialog = null;
    
    public MouseBehavior( View view ) {
        super(view);
        init();
        canvasCenter = new java.awt.geom.Point2D.Float();
    }
    
    /**
     * Canvas3D size will be calculated on MouseEvent.MOUSE_ENTERED
     * or component resize
     */
    
    public MouseBehavior() {
        init();
        canvasCenter = new java.awt.geom.Point2D.Float();
    }
     

    /**
      * Set the TransformGroup on which this behavior acts
      * This is called when the behavior is set using
      * VPBehaviorControl.setVPBehavior
      */
    public void setTarget( TransformGroup targetTG ) {
        this.targetTG = targetTG;
    }
    
    private void init() {
        
        // For some reason having MOUSE_MOTION and MOUSE events in the same wakeup
        // results in MOUSE events being lost
        WakeupOnAWTEvent awt1 = new WakeupOnAWTEvent( AWTEvent.MOUSE_MOTION_EVENT_MASK |
        AWTEvent.COMPONENT_EVENT_MASK );
        WakeupOnAWTEvent awt2 = new WakeupOnAWTEvent( AWTEvent.MOUSE_EVENT_MASK );
        WakeupOnAWTEvent awt3 = new WakeupOnAWTEvent(  AWTEvent.KEY_EVENT_MASK );
        WakeupOnElapsedFrames frameCondition = new WakeupOnElapsedFrames( 0 );

        bothCondition = new WakeupOr( new WakeupCriterion[] { awt1, awt2, awt3, frameCondition } );
        awtCondition = new WakeupOr( new WakeupCriterion[] { awt1, awt2, awt3 } );
    }

    public void initialize() {
        wakeupOn( awtCondition );
    }

    public void processStimulus( java.util.Enumeration enum ) {
        while(enum.hasMoreElements()) {
            WakeupCondition wakeup = (WakeupCondition)enum.nextElement();
            if (wakeup instanceof WakeupOnAWTEvent) {
                processAWTEvents( ((WakeupOnAWTEvent)wakeup).getAWTEvent() );
            } else if (wakeup instanceof WakeupOnElapsedFrames) {
                integrateTransforms();
            }
        }


        if (motion) {
            wakeupOn( bothCondition );
        } else {
            wakeupOn( awtCondition );
        }
    }

    private void processAWTEvents( java.awt.AWTEvent[] events ) {
        
        for( int i=0; i<events.length; i++) {
            if (events[i] instanceof MouseEvent) {
                //System.out.println( events[i] );
                if ((events[i].getID()==MouseEvent.MOUSE_DRAGGED ||
                    events[i].getID()==MouseEvent.MOUSE_PRESSED ||
                    events[i].getID()==MouseEvent.MOUSE_RELEASED) &&
                    !ignoreMouseMotion ) {
                    processMouseEvent( (MouseEvent)events[i] );
                } else if (events[i].getID()==MouseEvent.MOUSE_ENTERED ) {
                    Component component = ((MouseEvent)events[i]).getComponent();
                    if (component instanceof javax.media.j3d.Canvas3D) {
                        canvasBounds = component.getBounds( canvasBounds );
                        canvasCenter.x = canvasBounds.width/2;
                        canvasCenter.y = canvasBounds.height/2;
                        deadXSize = 15f / canvasCenter.x;
                        deadYSize = 15f / canvasCenter.y;  
                        if (!ignoreMouseMotion) {
                            component.setCursor( Cursor.getPredefinedCursor( Cursor.CROSSHAIR_CURSOR ));
                        }
                    }
                } else if (events[i].getID()==MouseEvent.MOUSE_EXITED ) {
                    Component component = ((MouseEvent)events[i]).getComponent();
                    if (component instanceof javax.media.j3d.Canvas3D) {
                        component.setCursor( Cursor.getDefaultCursor() );
                    }
                }
            } else if (events[i] instanceof KeyEvent) {
                int keyCode = ((KeyEvent)events[i]).getKeyCode();
                int id = ((KeyEvent)events[i]).getID();
                if (keyCode == KeyEvent.VK_ESCAPE && id==KeyEvent.KEY_RELEASED) {
                    ignoreMouseMotion = !ignoreMouseMotion;
                    Component component = ((KeyEvent)events[i]).getComponent();
                    if (ignoreMouseMotion) {
                        component.setCursor( Cursor.getDefaultCursor());
                    } else { 
                        component.setCursor( Cursor.getPredefinedCursor( Cursor.CROSSHAIR_CURSOR ));
                    }
                }
            } else if (events[i] instanceof ComponentEvent) {
                if (events[i].getID()==ComponentEvent.COMPONENT_RESIZED) {
                    canvasBounds = ((ComponentEvent)events[i]).getComponent().getBounds( canvasBounds );
                    canvasCenter.x = canvasBounds.width/2f;
                    canvasCenter.y = canvasBounds.height/2f;
                    deadXSize = 15f / canvasCenter.x;
                    deadYSize = 15f / canvasCenter.y;  
                }
            } //else
            //System.out.println("Unused event "+events[i]);

        }
    }

    /**
     * Process the mouse events, this is called every time an event
     * arrives
     */
    protected abstract void processMouseEvent( MouseEvent evt );
    
    /**
     * Called once per frame (if the view is moving) to calculate the new
     * view platform transform
     */
    protected abstract void integrateTransforms();

    /**
     * Called when the user wants to change the parameters of the
     * behavior.
     * 
     * This method may be called for the AWT event thread
     */
    public void editParameters( java.awt.Frame parent ) {
        if (flyControlDialog==null) {
            flyControlDialog = new FlyControlDialog( parent, true );
        }
        
        flyControlDialog.setStep( MAX_VELOCITY );
        flyControlDialog.setAngle( (float)Math.toDegrees(MAX_ANGLE) );
        flyControlDialog.setVisible( true );
        
        if (!flyControlDialog.userCancelled()) {
            MAX_ANGLE = (float)Math.toRadians( flyControlDialog.getAngle() );
            MAX_VELOCITY = flyControlDialog.getStep();
        }
    }
}

