/*
 *  @(#)VPDefaultCollision.java 1.8 01/01/09 10:39:30
 *
 * Copyright (c) 2000-2001 Sun Microsystems, Inc. All Rights Reserved.
 */
package com.sun.j3d.demos.utils.vpbehaviors;

import java.util.ArrayList;

import javax.media.j3d.BranchGroup;
import javax.media.j3d.Transform3D;
import javax.media.j3d.Locale;
import javax.media.j3d.Node;
import javax.media.j3d.Shape3D;

import javax.vecmath.Point3d;
import javax.vecmath.Vector3f;

import com.sun.j3d.utils.picking.PickTool;
import com.sun.j3d.utils.picking.PickResult;
import com.sun.j3d.demos.utils.scenegraph.traverser.TreeScan;
import com.sun.j3d.demos.utils.scenegraph.traverser.NodeChangeProcessor;


/**
 * A simple collision implementation that uses Picking to determine if
 * a collision will occurr.
 *
 * The rays of a SweptVolume are cast between the current viewPlatform and
 * next viewPlatform transform. Only the rays for the non-zero velocity 
 * directions are cast. For example if the VP is moving forward (-ve Z) then
 * frontRays are checked.
 *
 * This opimization means that this collision implementation is not appropriate for
 * checking for collision between the viewPlatform and fast moving animated geometry.
 * This feature will be added in a future version of the API.
 *
 * @author Paul Byrne
 * @version	1.8, 01/09/01
 */
public class VPDefaultCollision implements VPCollisionInterface{

    protected ArrayList branchGroups;
    protected PickTool pickTool=null;

    private Point3d segmentEnd = new Point3d();
    private Point3d segmentStart = new Point3d();
    
    protected SweptVolume sweptVolume;
     
    /**
     * Return minimal amount of information of collision
     * which is the Shape3D and the RayIndex and Direction
     */
    public static final int ALLOW_COLLISION_SHAPE = 0x00;
    
    /**
     * Return the point3d for each collision
     */
    public static final int ALLOW_COLLISION_POINT = 0x01;
    
    /**
     * Return the geometry for each collision
     */
    public static final int ALLOW_COLLISION_GEOMETRY = 0x02;
    
    private int mode;

    public VPDefaultCollision( int mode ) {
        this.mode = mode;
        branchGroups = new ArrayList();
    }
    
    /**
     * Creates a default VPDefaultCollision with a mode of
     * ALLOW_COLLISION
     */
    public VPDefaultCollision() {
        this( ALLOW_COLLISION_SHAPE );
    }

    /**
     * Add a BranchGroup to check for collision
     *
     * Current implementation only support a single BranchGroup
     * @see setCollisionLocale
     *
     * @param bg BranchGroup to add to set of BranchGroups checked for
     * collision
     */
    public void addCollisionBG( BranchGroup bg ) {
        if (pickTool==null) {
            pickTool = new PickTool( bg );
        } else {
            throw new RuntimeException( "Implementation only supports one CollisionBG");
        }

        branchGroups.add( bg );
        
        if (mode > ALLOW_COLLISION_SHAPE)
            pickTool.setMode( PickTool.GEOMETRY_INTERSECT_INFO );
        else
            pickTool.setMode( PickTool.GEOMETRY );
    }

    /**
     * Check for collision with all geometry in <code>locale</code>
     */
    public void setCollisionLocale(Locale locale) {
        pickTool = new PickTool( locale );
        if (mode > ALLOW_COLLISION_SHAPE)
            pickTool.setMode( PickTool.GEOMETRY_INTERSECT_INFO );
        else
            pickTool.setMode( PickTool.GEOMETRY );
    }

    /**
     * Get the number of collision BranchGroups that are checked for
     * collision
     */
    public int getCollisionBGCount() {
        return branchGroups.size();
    }

    /**
     * Get the BranchGroup at the specified index that is used for
     * collision detection
     */
    public BranchGroup getCollisionBG( int index ) {
        return (BranchGroup)branchGroups.get( index );
    }

    /**
     * Check for collision when casting the swept bounds from currentLocation to
     * nextLocation. The orientation of the swept bounds is set to the orientation at
     * nextLocation.
     *
     * This implementation only processes the first ray in each direction. It
     * also requires that a ray exists in each direction.
     *
     * @param currentLocation The current location of the view
     * @param nextLocation Will contain the next non collision location of the view on return
     * @param velocity The current velocity vector for the view
     * @param roll The change in roll for this frame
     * @param pitch The change in pitch for this frame
     * @param yaw The change in yaw for this frame
     */
    public SweptVolumeCollision getCollisions(Transform3D currentLocation,Transform3D nextLocation,Vector3f velocity,float roll,float pitch,float yaw) {
        
        boolean collision = false;
        SweptVolumeCollision ret = new SweptVolumeCollision();
        
        Transform3D yawT = new Transform3D();
        Transform3D pitchT = new Transform3D();
        Transform3D rollT = new Transform3D();
        Transform3D velocityT = new Transform3D();
        Transform3D destination = new Transform3D();
        
        yawT.rotY( yaw );
        pitchT.rotX( pitch );
        rollT.rotZ( roll );

        velocityT.set( velocity );
        velocityT.mul( yawT );
        velocityT.mul( pitchT );
        velocityT.mul( rollT );

        //System.out.println(velocity+"  "+acceleration);
        
        destination.set( currentLocation );

        destination.mul( yawT );
        destination.mul( pitchT );
        destination.mul( rollT );
        destination.mul( velocityT );
                
        collision |= checkZDirection( currentLocation, destination, velocity, ret );
        collision |= checkXDirection( currentLocation, destination, velocity, ret );
        collision |= checkYDirection( currentLocation, destination, velocity, ret );
                
        // If the volume is in collision return the original transform, ie
        // don't move.
        // An alternative approach would be to calculate the 'best' move from
        // the current location, ie move as far as possible without causing
        // collision.
        if (collision) {
            nextLocation.set( currentLocation );
            return null;
        } else {
            nextLocation.set( destination );
            return ret;
        }
    }
    
    protected boolean checkZDirection( Transform3D currentLocation, Transform3D destination, Vector3f velocity,
                                     SweptVolumeCollision ret ) {
        boolean collision = false;
        if (velocity.z<0f) {   // Check on Z axis
            PickResult result = checkDirection( currentLocation, destination, sweptVolume.frontRays[0], sweptVolume.frontRays[1], velocity.z * -1f );
            if (result!=null) {
                collision |= true;
                RayCollision ray = new RayCollision();
                ray.rayDirection = RayCollision.FRONT;
                ray.rayIndex = 0;
                ray.collisionShape = (Shape3D)result.getObject();
                if (mode>ALLOW_COLLISION_SHAPE) {
                    ray.collisionGeometryArray = result.getIntersection(0).getGeometryArray();
                    ray.collisionPoint = result.getIntersection(0).getPointCoordinatesVW();
                }
                ret.addRayCollision( ray );
            }
        } else if (velocity.z>0f) {
            PickResult result = checkDirection( currentLocation, destination, sweptVolume.backRays[0], sweptVolume.backRays[1], velocity.z );
            if (result!=null) {
                collision |= true;
                RayCollision ray = new RayCollision();
                ray.rayDirection = RayCollision.BACK;
                ray.rayIndex = 0;
                ray.collisionShape = (Shape3D)result.getObject();
                if (mode>ALLOW_COLLISION_SHAPE) {
                    ray.collisionGeometryArray = result.getIntersection(0).getGeometryArray();
                    ray.collisionPoint = result.getIntersection(0).getPointCoordinatesVW();
                }
                ret.addRayCollision( ray );
            }
        }

        return collision;
    }
    
    protected boolean checkXDirection( Transform3D currentLocation, Transform3D destination, Vector3f velocity,
                                     SweptVolumeCollision ret ) {
        boolean collision = false;
        if (velocity.x>0f) {   // Check on X axis
            PickResult result = checkDirection( currentLocation, destination, sweptVolume.rightRays[0], sweptVolume.rightRays[1], velocity.x );
            if (result!=null) {
                collision |= true;
                RayCollision ray = new RayCollision();
                ray.rayDirection = RayCollision.RIGHT;
                ray.rayIndex = 0;
                ray.collisionShape = (Shape3D)result.getObject();
                if (mode>ALLOW_COLLISION_SHAPE) {
                    ray.collisionGeometryArray = result.getIntersection(0).getGeometryArray();
                    ray.collisionPoint = result.getIntersection(0).getPointCoordinatesVW();
                }
                ret.addRayCollision( ray );
            }
        } else if (velocity.x<0f) {
            PickResult result = checkDirection( currentLocation, destination, sweptVolume.leftRays[0], sweptVolume.leftRays[1], velocity.x * -1f );
            if (result!=null) {
                collision |= true;
                RayCollision ray = new RayCollision();
                ray.rayDirection = RayCollision.LEFT;
                ray.rayIndex = 0;
                ray.collisionShape = (Shape3D)result.getObject();
                if (mode>ALLOW_COLLISION_SHAPE) {
                    ray.collisionGeometryArray = result.getIntersection(0).getGeometryArray();
                    ray.collisionPoint = result.getIntersection(0).getPointCoordinatesVW();
                }
                ret.addRayCollision( ray );
            }
        }
        
        return collision;                                     
    }
    
    protected boolean checkYDirection( Transform3D currentLocation, Transform3D destination, Vector3f velocity,
                                     SweptVolumeCollision ret ) {
        boolean collision = false;
        if (velocity.y>0f) {   // Check on Y axis
            PickResult result = checkDirection( currentLocation, destination, sweptVolume.upRays[0], sweptVolume.upRays[1], velocity.y );
            if (result!=null) {
                collision |= true;
                RayCollision ray = new RayCollision();
                ray.rayDirection = RayCollision.UP;
                ray.rayIndex = 0;
                ray.collisionShape = (Shape3D)result.getObject();
                if (mode>ALLOW_COLLISION_SHAPE) {
                    ray.collisionGeometryArray = result.getIntersection(0).getGeometryArray();
                    ray.collisionPoint = result.getIntersection(0).getPointCoordinatesVW();
                }
                ret.addRayCollision( ray );
            }
        } else if (velocity.y<0f) {
            PickResult result = checkDirection( currentLocation, destination, sweptVolume.downRays[0], sweptVolume.downRays[1], velocity.y * -1f );
            if (result!=null) {
                collision |= true;
                RayCollision ray = new RayCollision();
                ray.rayDirection = RayCollision.DOWN;
                ray.rayIndex = 0;
                ray.collisionShape = (Shape3D)result.getObject();
                if (mode>ALLOW_COLLISION_SHAPE) {
                    ray.collisionGeometryArray = result.getIntersection(0).getGeometryArray();
                    ray.collisionPoint = result.getIntersection(0).getPointCoordinatesVW();
                }
                ret.addRayCollision( ray );
            }
        }
    
        return collision;
    }

    protected PickResult checkDirection( Transform3D currentLocation, Transform3D nextLocation, 
                                        Vector3f rayStart, Vector3f rayEnd, float stepSize ) {
                                            
        if (stepSize<1f) stepSize =1f;
        
        segmentEnd.x = rayEnd.x * stepSize;
        segmentEnd.y = rayEnd.y * stepSize;
        segmentEnd.z = rayEnd.z * stepSize;
        segmentStart.x = rayStart.x;
        segmentStart.y = rayStart.y;
        segmentStart.z = rayStart.z;
        
        nextLocation.transform( segmentEnd );
        currentLocation.transform( segmentStart );
        
        //javax.vecmath.Vector3d loc = new javax.vecmath.Vector3d();
        //currentLocation.get( loc );
        //System.out.println("Location "+loc);
        //System.out.println("Seg "+segmentStart+"  "+segmentEnd);
        
        pickTool.setShapeSegment( segmentStart, segmentEnd );
        return pickTool.pickClosest();
    }
        
    /**
      * Set the capabilities in the BranchGroup to allow this collision 
      * detection to work
      */
    public void setCapabilities( BranchGroup branchGroup ) {
        final int m = this.mode;
        NodeChangeProcessor proc = new NodeChangeProcessor() {
            public void changeNode( Node node ) {
                if (m > ALLOW_COLLISION_SHAPE)
                    pickTool.setCapabilities( node, PickTool.INTERSECT_FULL );
                else
                    pickTool.setCapabilities( node, PickTool.INTERSECT_TEST );
            }
        };

        TreeScan.findNode( branchGroup, new Class[] { javax.media.j3d.Shape3D.class,
            javax.media.j3d.Morph.class }, proc, false, true );
    }
    
    /** Set the SweptsweptVolume to be used in collision calculations
    */
    public void setVPSweptVolume(SweptVolume volume) {
        sweptVolume = volume;
    }
    
    /** Get the SweptVolume used in collision calculations
    */
    public SweptVolume getVPSweptVolume() {
        return sweptVolume;
    }
    
}
