/*
 *  @(#)VPDriveCollision.java 1.2 01/01/09 10:39:48
 *
 * Copyright (c) 2000-2001 Sun Microsystems, Inc. All Rights Reserved.
 */
package com.sun.j3d.demos.utils.vpbehaviors;

import javax.media.j3d.Transform3D;
import javax.vecmath.Vector3f;
import javax.vecmath.Vector3d;
import javax.vecmath.Point3d;
import com.sun.j3d.utils.picking.PickResult;

/**
 * A collision implementation that places the view a set distance above
 * the 'floor' and then applies collision detection.
 *
 * The floor is determined as the first geometry that intersects the ray
 * (0,large +ve, 0) -> ( 0, -1, 0 )
 *
 * @author  Paul Byrne
 * @version 1.2 01/09/01
 */
public class VPDriveCollision extends VPDefaultCollision {

    private Point3d rayStart = new Point3d( 0, 1000, 0 );
    private static final Vector3d rayVec = new Vector3d( 0, -1, 0 );
    private Vector3f pos = new Vector3f();
    private float height = 1.2f;
    
    /** Creates new VPDriveCollision */
    public VPDriveCollision(int mode) {
        super(mode);
    }
    
    /** Creates new VPDriveCollision */
    public VPDriveCollision() {
        super();
    }
    
    /**
     * Set the height the viewer will be placed above the floor
     */
    public void setHeight( float height ) {
        this.height = height;
    }
    
    /**
     * Get the height the viewer is from the floor
     */
    public float getHeight() {
        return height;
    }
    
    /**
     * Check for collision when casting the swept bounds from currentLocation to
     * nextLocation. The orientation of the swept bounds is set to the orientation at
     * nextLocation.
     *
     * Also force the view to be the specified height above the floor.
     *
     * This implementation only processes the first ray in each direction. It
     * also requires that a ray exists in each direction.
     *
     * @param currentLocation The current location of the view
     * @param nextLocation Will contain the next non collision location of the view on return
     * @param velocity The current velocity vector for the view
     * @param roll The change in roll for this frame
     * @param pitch The change in pitch for this frame
     * @param yaw The change in yaw for this frame
     */
    public SweptVolumeCollision getCollisions(Transform3D currentLocation,Transform3D nextLocation,Vector3f velocity,float roll,float pitch,float yaw) {
        
        boolean collision = false;
        SweptVolumeCollision ret = new SweptVolumeCollision();
        
        Transform3D yawT = new Transform3D();
        Transform3D pitchT = new Transform3D();
        Transform3D rollT = new Transform3D();
        Transform3D velocityT = new Transform3D();
        Transform3D destination = new Transform3D();
        
        yawT.rotY( yaw );
        pitchT.rotX( pitch );
        rollT.rotZ( roll );

        velocityT.set( velocity );
        velocityT.mul( yawT );
        velocityT.mul( pitchT );
        velocityT.mul( rollT );

        destination.set( currentLocation );

        destination.mul( yawT );
        destination.mul( pitchT );
        destination.mul( rollT );
        destination.mul( velocityT );
        
        // Find the floor
        destination.get( pos );
        rayStart.x = pos.x;
        rayStart.z = pos.z;
        rayStart.y = 100000;
        
        pickTool.setShapeRay( rayStart, rayVec );
        PickResult floor = pickTool.pickClosest();
        if (floor!=null) {   
            pos.y = (float)floor.getIntersection(0).getPointCoordinates().y+height;
            destination.setTranslation( pos );
        }
        
        collision |= checkZDirection( currentLocation, destination, velocity, ret );
        collision |= checkXDirection( currentLocation, destination, velocity, ret );
        collision |= checkYDirection( currentLocation, destination, velocity, ret );
        
        // In many cases the checkYDirection call can be omitted. However it is 
        // included in this general class to check for collision above the
        // view position.
                
        // If the volume is in collision return the original transform, ie
        // don't move.
        // An alternative approach would be to calculate the 'best' move from
        // the current location, ie move as far as possible without causing
        // collision.
        if (collision) {
            nextLocation.set( currentLocation );
            return null;
        } else {
            nextLocation.set( destination );
            return ret;
        }
    }
}
